package io.gitlab.jfronny.inceptum.gtk.menu;

import io.gitlab.jfronny.commons.throwable.ThrowingRunnable;
import io.gitlab.jfronny.inceptum.common.Utils;
import io.gitlab.jfronny.inceptum.gtk.util.I18n;
import org.gtk.gio.MenuItem;
import org.gtk.gio.*;
import org.gtk.glib.Variant;
import org.gtk.glib.VariantType;
import org.gtk.gtk.Application;
import org.gtk.gtk.PopoverMenu;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.function.BiFunction;
import java.util.function.Consumer;

public class MenuBuilder {
    private static final Object LOCK = new Object();

    private static Menu getRootMenu(Application app) {
        synchronized (LOCK) {
            var currentMenu = app.menubar;
            if (currentMenu == null) {
                var menu = new Menu();
                app.menubar = menu;
                return menu;
            } else {
                return Menu.castFrom(currentMenu);
            }
        }
    }

    private final ActionMap map;
    private final Map<String, Action> refs = new LinkedHashMap<>();
    private final Menu menu;
    private final String prefix;

    public MenuBuilder(Application app) {
        this(app, getRootMenu(app), "");
    }

    public MenuBuilder(ActionMap map, Menu menu, String prefix) {
        if (!prefix.isEmpty() && !prefix.endsWith(".")) prefix += ".";
        this.map = Objects.requireNonNull(map);
        this.menu = Objects.requireNonNull(menu);
        this.prefix = Objects.requireNonNull(prefix);
    }

    public ButtonItem button(String name, ThrowingRunnable<?> onClick) {
        return literalButton(name, I18n.get("menu." + prefix + name), onClick);
    }

    public ButtonItem literalButton(String internalName, String label, ThrowingRunnable<?> onClick) {
        internalName = prefix + internalName;
        SimpleAction sAct = new SimpleAction(internalName, null);
        addAction(internalName, sAct);
        sAct.onActivate(($, variant) -> {
            try {
                onClick.run();
            } catch (Throwable e) {
                Utils.LOGGER.error("Could not execute action", e);
            }
        });
        menu.appendItem(new MenuItem(label, "app." + internalName));
        return new ButtonItem(sAct);
    }

    public ToggleItem toggle(String name, boolean initial, Consumer<Boolean> action) {
        name = prefix + name;
        SimpleAction sAct = SimpleAction.newStateful(name, null, Variant.newBoolean(initial));
        addAction(name, sAct);
        sAct.onActivate(($, variant) -> {
            boolean state = !sAct.getState().getBoolean();
            sAct.state = Variant.newBoolean(state);
            action.accept(state);
        });
        menu.appendItem(new MenuItem(I18n.get("menu." + name), "app." + name));
        return new ToggleItem(sAct);
    }

    public <T> RadioItem<T> radio(String name, T initial, List<T> options, Consumer<T> action) {
        return literalRadio(name, initial, options, (i, t) -> I18n.get("menu." + name, i), action);
    }

    public <T> RadioItem<T> literalRadio(String name, T initial, List<T> options, BiFunction<Integer, T, String> stringifier, Consumer<T> action) {
        Objects.requireNonNull(options);
        name = prefix + name;
        SimpleAction sAct = SimpleAction.newStateful(name, new VariantType("i"), Variant.newInt32(options.indexOf(initial)));
        addAction(name, sAct);
        sAct.onActivate(($, variant) -> {
            sAct.state = variant;
            action.accept(options.get(variant.getInt32()));
        });
        int i = 0;
        for (T option : options) {
            menu.appendItem(new MenuItem(stringifier.apply(i, option), "app." + name + "(" + i + ")"));
            i++;
        }
        return new RadioItem<>(sAct, options);
    }

    public MenuBuilder submenu(String name) {
        name = prefix + name;
        Menu submenu = new Menu();
        menu.appendSubmenu(I18n.get("menu." + name), submenu);
        return new MenuBuilder(map, submenu, name);
    }

    public void clear() {
        menu.removeAll();
        refs.forEach((name, action) -> {
            map.removeAction(name);
        });
        refs.clear();
    }

    private void addAction(String name, SimpleAction action) {
        map.addAction(action);
        refs.put(name, action);
    }

    public Menu getMenu() {
        return menu;
    }

    public PopoverMenu asPopover() {
        return PopoverMenu.newFromModel(menu);
    }
}
